﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  AVR Chipspezifische Basisfunktionalitäten                                           //
//  _________________________________  													//
//                                                                                      //
//  Copyright (c) 2010 Reimar Grasbon													//
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#pragma once

#include <avr/io.h>

//////////////////////////////////////////////////////////////////////////////////////////
//  Port-Konfiguration  //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

#define _T_  0  // Eingabe Tristate
#define _P_  1  // Eingabe Pullup
#define _L_  2  // Ausgabe Low
#define _H_  3  // Ausgabe High

#define CONCATENATE_DDR(port)  (DDR  ## port)
#define CONCATENATE_OUTP(port) (PORT ## port)
#define CONCATENATE_INP(port)  (PIN  ## port)

#define ConfPort(port, cfg0, cfg1, cfg2, cfg3, cfg4, cfg5, cfg6, cfg7)   \
do                                                                       \
{   CONCATENATE_DDR(port) =  (((cfg0&0x02)>>1)<<0) |                     \
                             (((cfg1&0x02)>>1)<<1) |                     \
                             (((cfg2&0x02)>>1)<<2) |                     \
                             (((cfg3&0x02)>>1)<<3) |                     \
                             (((cfg4&0x02)>>1)<<4) |                     \
                             (((cfg5&0x02)>>1)<<5) |                     \
                             (((cfg6&0x02)>>1)<<6) |                     \
                             (((cfg7&0x02)>>1)<<7);                      \
    CONCATENATE_OUTP(port) = (( cfg0&0x01)    <<0) |                     \
                             (( cfg1&0x01)    <<1) |                     \
                             (( cfg2&0x01)    <<2) |                     \
                             (( cfg3&0x01)    <<3) |                     \
                             (( cfg4&0x01)    <<4) |                     \
                             (( cfg5&0x01)    <<5) |                     \
                             (( cfg6&0x01)    <<6) |                     \
                             (( cfg7&0x01)    <<7);                      \
} while (0)

//////////////////////////////////////////////////////////////////////////////////////////
//  Port-Pin Konfiguration  //////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

// Schreiben
#define ConfPortPin(port, pin, cfg)                                                           \
do                                                                                            \
{   CONCATENATE_DDR(port)  = (CONCATENATE_DDR(port)  & ~(1<<(pin))) | (((cfg&0x02)>>1)<<pin); \
    CONCATENATE_OUTP(port) = (CONCATENATE_OUTP(port) & ~(1<<(pin))) | (( cfg&0x01)    <<pin); \
} while (0)

#define ConfPin(portPin, cfg)    ConfPortPin(portPin, cfg)

// Lesen
#define PortPinConf(port, pin)  ((!!(CONCATENATE_DDR(port)&(1<<(pin)))<<1) | \
                                  !!(CONCATENATE_OUTP(port)&(1<<(pin))))

#define PinConf(portPin)        PortPinConf(portPin)

//////////////////////////////////////////////////////////////////////////////////////////
//  Port Zugriff  ////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
#define WritePort(port, byte) CONCATENATE_OUTP(port) = (byte)
#define ReadPort(port)        CONCATENATE_INP(port)

//////////////////////////////////////////////////////////////////////////////////////////
//  Port-Pin Zugriff  ////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
#define SetPortPin(port, pin)      (CONCATENATE_OUTP(port) |=  (1<<(pin)))
#define ClrPortPin(port, pin)      (CONCATENATE_OUTP(port) &= ~(1<<(pin)))
#define InvPortPin(port, pin)      (CONCATENATE_OUTP(port) ^=  (1<<(pin)))
#define GetPortPin(port, pin)      (CONCATENATE_INP(port)   &  (1<<(pin)))
#define ChgPortPin(port, pin, set) (set ? SetPortPin(port, pin) : ClrPortPin(port, pin))

#define SetPin(portPin)      SetPortPin(portPin)
#define ClrPin(portPin)      ClrPortPin(portPin)
#define InvPin(portPin)      InvPortPin(portPin)
#define GetPin(portPin)      GetPortPin(portPin)
#define ChgPin(portPin, set) ChgPortPin(portPin, set)

//////////////////////////////////////////////////////////////////////////////////////////
//  Unmittelbarer Sprung zu Speicheradresse  /////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
#ifdef __AVR_ATmega8__
    #define GOTO(adr) __asm__ volatile ("ijmp"::"z" ((adr)/2) )
#else
    #define GOTO_REDIRECT_ADR_TO_STRING(s) #s
    #define GOTO(adr) __asm__ volatile ("jmp "GOTO_REDIRECT_ADR_TO_STRING(adr))
#endif

//////////////////////////////////////////////////////////////////////////////////////////
//  Kritischer Abschnitt (Atomare Operation)  ////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static __inline__ void EnableInterrupts(void)  { __asm__ __volatile__ ("sei" ::); }
static __inline__ void DisableInterrupts(void) { __asm__ __volatile__ ("cli" ::); }
static __inline__ uint8_t __iCli(void)
{
    DisableInterrupts();
    return 1;
}
static __inline__ void   __iSei(const uint8_t* v)
{
    if (((*v) & (1<<SREG_I))) EnableInterrupts();
    __asm__ volatile ("" ::: "memory");
}
#define Atomic for (uint8_t _sreg_save_ __attribute__((__cleanup__(__iSei))) = SREG,\
                    _it_ = __iCli(); _it_ != 0 ; _it_ = 0)


//////////////////////////////////////////////////////////////////////////////////////////
//  Freeze10us(): Aktive Warteschleife. Verzögert um ca. 10µs. Die Anzahl der Schlei-   //
//  ------------  fendurchläufe wird auf Grundlage der Prozessortaktrate in <F_CPU>     //
//  ermittelt, deren Wert von 500 kHz bis 40 Mhz (bzw. 4 MHz bis 300 Mhz wenn <CKDIV8>  //
//  gesetzt ist) reichen darf.                                                          //
//  Man beachte: Der zum Aufruf der Funktion erzeugte Overhead des Compilers wird in    //
//  der Rechnung nicht berücksichtigt so dass die Laufzeit der Funktion geringfügig     //
//  länger sein wird.                                                                   //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static __inline__ void Freeze10us(void)
{
    uint16_t  i = 10*(F_CPU/(1+CKDIV8*7)/1000)/5000;       // 10 => #µs

    __asm__ volatile ("1: sbiw %0,1 \n\t"
                      "nop          \n\t"
                      "brne 1b"
                      : "=w" (i)
                      : "0" (i)
                      );
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Freeze100us(): Aktive Warteschleife. Verzögert um ca. 100µs. Die Anzahl der Schlei- //
//  -------------  fendurchläufe wird auf Grundlage der Prozessortaktrate in <F_CPU>    //
//  ermittelt, deren Wert von 500 kHz bis 40 Mhz (bzw. 4 MHz bis 300 Mhz wenn <CKDIV8>  //
//  gesetzt ist) reichen darf.                                                          //
//  Man beachte: Der zum Aufruf der Funktion erzeugte Overhead des Compilers wird in    //
//  der Rechnung nicht berücksichtigt so dass die Laufzeit der Funktion geringfügig     //
//  länger sein wird.                                                                   //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static __inline__ void Freeze100us(void)
{
    uint16_t  i = 100*(F_CPU/(1+CKDIV8*7)/1000)/5000;       // 100 => #µs

    __asm__ volatile ("1: sbiw %0,1 \n\t"
                      "nop          \n\t"
                      "brne 1b"
                      : "=w" (i)
                      : "0" (i)
                      );
}

//////////////////////////////////////////////////////////////////////////////////////////

