﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Tonsignale und Tonabfolgen                                                          //
//  _____________________________________________________                               //
//                                                                                      //
//  Copyright (c) 2017 Reimar Grasbon													//
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include <avr/pgmspace.h>
#include "Board.h"
#include "ToneSignal.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Definitionen  ////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
#define TONESEQUENCE_END    { (Note)255, 0 }  // Kontrollcode für: Tonabfolge beenden	
#define TONESEQUENCE_REPEAT { (Note)255, 1 }  // Kontrollcode für: Tonabfolge wiederholen

//////////////////////////////////////////////////////////////////////////////////////////
//  Typen  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
typedef struct              // Ton:
{   Note    pitch;          // · Tonhöhe
    UInt08  sustain;        // · Haltedauer: 1 .. 250 [10 ms]
} Tone;

//////////////////////////////////////////////////////////////////////////////////////////
//  Tonabfolgen  /////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
const Tone  sequence00[] PROGMEM =  {   { NOTE_C6, 20 }, { NOTE_XX,  20 },
                                        { NOTE_E6,  5 },
                                        { NOTE_G6,  5 },
                                        { NOTE_C7, 25 }, TONESEQUENCE_END
                                        };
										
//////////////////////////////////////////////////////////////////////////////////////////
const Tone  sequence01[] PROGMEM =  {   { NOTE_A6,  5 },
										{ NOTE_A8,  5 }, TONESEQUENCE_END
                                        };
										
//////////////////////////////////////////////////////////////////////////////////////////
const Tone  sequence02[] PROGMEM =  {   { NOTE_A5,  5 },
										{ NOTE_A7,  5 }, TONESEQUENCE_END
                                        };

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static ToneSignal   currentSignal = TONESIGNAL_NONE;
static const Tone*  sequence      = NULL;

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  SequenceOf(): Liefert die zum angegebenen Tonesignal passende Tonsequenz oder NULL,	//
//  ------------  falls 'signal' TONESIGNAL_NONE ist.									//
//                                                                                      //
//  Eingabe: · signal: Das Tonsignal, dessen Tonsequenz ermittelt werden soll als Auf-	//
//					   zählungstyp 'ToneSignal'.										//
//  Ausgabe: Zeiger auf die entsprechende Tonsequenz oder NULL.							//
//////////////////////////////////////////////////////////////////////////////////////////
const Tone*
SequenceOf(ToneSignal signal)
{	
	static const Tone* const table[] PROGMEM =
	{   sequence00, sequence01, sequence02
	};	    
	if ((UInt08)signal > CARDINALITY(table) || signal == TONESIGNAL_NONE) return NULL;
	return (const Tone*)pgm_read_word(&table[(UInt08)signal-1]);	
}

//////////////////////////////////////////////////////////////////////////////////////////
//  IsContinuous(): Prüft, ob das angegebenen Tonesignal in einer Endlosschleife immer	//
//  --------------	wiederholt wird. Wird TONESIGNAL_NONE übergeben, liefert die		//
//	Funktion 'false' zurück.															//
//                                                                                      //
//  Eingabe: · signal: Das zu prüfende Signal als Aufzählungstyp 'ToneSignal'.          //
//  Ausgabe: 'true', wenn 'signal' eine endlos laufende Signalausgabe ist.				//
//////////////////////////////////////////////////////////////////////////////////////////
bool
IsContinuous(ToneSignal signal)
{
	const Tone* sequence = SequenceOf(signal);
	if (sequence == NULL) return false;
	while (true)
	{	Note   pitch   = pgm_read_byte(&(sequence->pitch));
		UInt08 sustain = pgm_read_byte(&(sequence->sustain));
		if (pitch == (Note)255)
		{	switch (sustain)
			{	case 0:  return false;
				case 1:  return true;
				default: ;
			}
		}
		sequence++;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
// Funktionen ////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  TONESIGNAL_Start(): Beginnt mit der Wiedergabe des angegebenen Tonesignals. Eine  	//
//  ------------------	gegebenenfalls laufende Signalausgabe wird beendet.				//
//                                                                                      //
//  Eingabe: · signal: Das auszugebende Signal als Aufzählungstyp 'ToneSignal'.         //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
TONESIGNAL_Start(ToneSignal signal)
{	
    BOARD_PlayTone(NOTE_XX, 0);    	
	sequence = SequenceOf(signal);
	if (sequence == NULL) currentSignal = TONESIGNAL_NONE;
	else				  currentSignal = signal;	
}

//////////////////////////////////////////////////////////////////////////////////////////
//  TONESIGNAL_Stop(): Beendet die gerade laufende Signalausgabe.						//
//  -----------------																	//
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
TONESIGNAL_Stop(void)
{    
	TONESIGNAL_Start(TONESIGNAL_NONE);	
}

//////////////////////////////////////////////////////////////////////////////////////////
//  TONESIGNAL_Signal(): Liefert das aktuell wiedergegebene Tonesignal oder TONE-		//
//  -------------------  SIGNAL_NONE, wenn gerade keine Signalausgabe stattfindet.		//
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
ToneSignal
TONESIGNAL_Current(void)
{
	return currentSignal;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Prozessaufrufe  //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
TONESIGNAL_Startup(void)
{
}

//////////////////////////////////////////////////////////////////////////////////////////
void
TONESIGNAL_Proceed(void)
{
	if (currentSignal == TONESIGNAL_NONE || !BOARD_IsSilence()) return;
	Note   pitch   = pgm_read_byte(&(sequence->pitch));
	UInt08 sustain = pgm_read_byte(&(sequence->sustain));
	if (pitch == (Note)255)
	{	switch (sustain)
		{	case 0:  TONESIGNAL_Stop();              break;
			case 1:  TONESIGNAL_Start(currentSignal); break;
			default: ;
		}
	}
	else	
	{	BOARD_PlayTone(pitch, sustain);
		sequence++;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
void
TONESIGNAL_Shutdown(void)
{
	TONESIGNAL_Stop();
}

//////////////////////////////////////////////////////////////////////////////////////////